"""
Simple script to test trt engine output vs. tensorflow pb output to provide to NVIDIA
Dependencies are kept minimal

Conf:
- python 2.7
- tensorflow 1.11
- tensorrt-5.0.0.10
- cuda-9.0
- cudnn7.3

Running on a P100 GPU

IMPORTANT:
As I am NOT able to run tensorflow then the trt engine builder in the same python process (else I get memory erorrs),
the workflow here is:
1) Create reference outputs using tensorflow: the output will be saved in a np file
    python test_trt_for_nvidia.py -o run_tf

2) Build trt engine
    python test_trt_for_nvidia.py -o run_tf

3) Run trt engine
    python test_trt_for_nvidia.py run_trt
"""

import subprocess

import argparse
import numpy as np
import tensorflow as tf
# noinspection PyUnresolvedReferences
import pycuda.driver as cuda
# noinspection PyUnresolvedReferences
import pycuda.autoinit
# noinspection PyUnresolvedReferences
import tensorrt as trt

__author__ = 'pbp'
__date__ = '10/2018'
__copyright__ = "Airbus Defence and Space"


# Parameters and constants definition
tf_pb_file = 'best_deploy.pb'
uff_file = 'best_deploy.uff'
trt_engine_file = 'best_deploy.trt_engine'
test_input_file = 'input_test.npy'
ref_output_file_tf = 'ref_output_from_tf'

input_node_name = 'ImagePatches'
output_node_name = 'Sum'
max_workspace_size_gb = 4
max_batch_size = 50
input_patch_size = 80


class HostDeviceMem(object):
    """
    Helper class to map between host (cpu) and device (gpu) memory
    """
    def __init__(self, host_mem, device_mem):
        self.host = host_mem
        self.device = device_mem

    def __str__(self):
        return "Host:\n" + str(self.host) + "\nDevice:\n" + str(self.device)

    def __repr__(self):
        return self.__str__()


def allocate_buffers(engine):
    """
    Helper function to allocate all buffers required for an engine, i.e. host/device inputs/outputs.

    Args:
        engine (tensorrt.ICudaEngine): input engine
    """
    # Init
    inputs, outputs, bindings, stream = [], [], [], cuda.Stream()

    # Create bindings
    for binding in engine:
        size = trt.volume(engine.get_binding_shape(binding))
        dtype = trt.nptype(engine.get_binding_dtype(binding))

        # Allocate host and device buffers
        host_mem = cuda.pagelocked_empty(size, dtype)
        device_mem = cuda.mem_alloc(host_mem.nbytes)

        # Append the device buffer to device bindings.
        bindings.append(int(device_mem))

        # Append to the appropriate list.
        if engine.binding_is_input(binding):
            inputs.append(HostDeviceMem(host_mem, device_mem))
        else:
            outputs.append(HostDeviceMem(host_mem, device_mem))

    return inputs, outputs, bindings, stream


def run_reference_tensorflow():
    """
    Run simple test to get reference output from tensorflow pb graph

    Note that I have to do save the result in a .npy file because if **in the same python process** I run tensorflow,
    and then I try to build / use the trt engine, then I get :
       [TensorRT] ERROR: runtime.cpp (24) - Cuda Error in allocate: 2
    """
    test_input = np.load(test_input_file)

    # Reload graph
    with tf.gfile.GFile(tf_pb_file, "rb") as f:
        graph_def = tf.GraphDef()
        graph_def.ParseFromString(f.read())

    with tf.Graph().as_default() as graph:
        tf.import_graph_def(graph_def, return_elements=[output_node_name])
        input_node = graph.get_tensor_by_name('import/ImagePatches:0')
        output_node = graph.get_tensor_by_name('import/Sum:0')

    # Run on test input and save
    with tf.Session(graph=graph) as sess:
        tf_output = sess.run(output_node, feed_dict={input_node: test_input})

    np.save(ref_output_file_tf, tf_output)


def build_trt_engine():
    """
    Build trt engine from .pb graph
    """
    # Convert to uff
    TRT_LOGGER = trt.Logger(trt.Logger.INFO)
    TRT_LOGGER.log(trt.Logger.INFO, 'Creating uff graph')
    subprocess.check_call(['convert-to-uff', '-o', uff_file, tf_pb_file])

    # Create trt engine
    TRT_LOGGER.log(trt.Logger.INFO,
                   'Creating trt engine with max_workspace_size={0}Gb, max_batch_size={1}'.format(
                       max_workspace_size_gb,
                       max_batch_size))

    with trt.Builder(TRT_LOGGER) as builder, builder.create_network() as network, trt.UffParser() as parser:
        # Max workspace size and batch size
        builder.max_workspace_size = max_workspace_size_gb * 1 << 30
        builder.max_batch_size = max_batch_size

        # Define input shape as NCHW
        input_shp = [max_batch_size] + [3, input_patch_size, input_patch_size]

        # Parse the Uff Network
        parser.register_input(input_node_name, input_shp, trt.UffInputOrder.NHWC)
        parser.register_output(output_node_name)
        parser.parse(uff_file, network)

        # Build engine
        engine = builder.build_cuda_engine(network)

        # Serialize engine
        if engine is not None:
            with open(trt_engine_file, 'wb') as stream:
                stream.write(engine.serialize())
        else:
            TRT_LOGGER.log(trt.Logger.ERROR, 'No engine created, something went wrong: check your logs')
            raise MemoryError('No engine created')


def run_trt_engine():
    """
    Run trt engine and compares results to ref tensorflow output
    """
    test_input = np.load(test_input_file)
    TRT_LOGGER = trt.Logger(trt.Logger.WARNING)

    # De-serialize engine
    with open(trt_engine_file, 'rb') as stream:
        serialized_engine = stream.read()

    # Build an engine, allocate buffers and create a stream.
    with trt.Runtime(TRT_LOGGER) as runtime:
        with runtime.deserialize_cuda_engine(serialized_engine) as engine:
            # Create buffers
            inputs, outputs, bindings, stream = allocate_buffers(engine)

            # Get expected input shape for the tensorrt engine (convention NCHW)
            # Note: this is not mandatory for this example but I do this in "production" mode as before running the
            # network I do pre-processing on my input images (slicing, padding...) so this enables to parameterize
            # my functions to be able to pass patches of correct dimensions
            expected_input_shape = engine.get_binding_shape(0)
            _max_batch_size = expected_input_shape[0]
            patch_size = (expected_input_shape[2], expected_input_shape[3], expected_input_shape[1])
            n_items_in = _max_batch_size * patch_size[0] * patch_size[1] * patch_size[2]

            with engine.create_execution_context() as context:
                # Copy input to page-locked host memory
                inputs_flatten = test_input.ravel()
                if inputs_flatten.shape[0] < n_items_in:
                    inputs_flatten = np.pad(inputs_flatten, (0, n_items_in - inputs_flatten.shape[0]), 'constant')
                np.copyto(inputs[0].host, inputs_flatten)

                # Transfer input data to the GPU.
                [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]

                # Run inference.
                context.execute_async(bindings=bindings, stream_handle=stream.handle)

                # Transfer predictions back from the GPU.
                [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]

                # Synchronize the stream
                stream.synchronize()

                # Return only the host outputs.
                [trt_output] = [out.host for out in outputs]
                print "Sample output (first 2 elements):"
                print trt_output[0:2]

    # Reshape
    # Test reshaping considering output is NCHW -> not better :-(
    #trt_output = trt_output.reshape((-1, patch_size[2], patch_size[0], patch_size[1]))
    #trt_output = np.moveaxis(trt_output, 1, -1)

    # Reshaping assuming output is NHWC -> honestly no idea
    trt_output = trt_output.reshape((-1, patch_size[0], patch_size[1], patch_size[2]))

    # Compare with ref
    ref_from_tf = np.load(ref_output_file_tf + '.npy')

    print "Ref from tensorflow (by band):"
    print ref_from_tf[0, 0:2, 0:2, 0]
    print ref_from_tf[0, 0:2, 0:2, 1]
    print ref_from_tf[0, 0:2, 0:2, 2]

    print "Output from TRT (by band):"
    print trt_output[0, 0:2, 0:2, 0]
    print trt_output[0, 0:2, 0:2, 1]
    print trt_output[0, 0:2, 0:2, 2]


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Simple script to compare outputs of tensorflow and outputs of trt on same reference inputs. '
                    'Weirdly, I am not able to perform 1) tensorflow execution 2) trt engine creation in the same '
                    'python process or I get "ERROR: runtime.cpp (24) - Cuda Error in allocate: 2". This is why'
                    'I separated three options')

    parser.add_argument('-o',
                        type=str,
                        help='This is either run_tf OR build_trt OR run_trt')

    # Parse args
    args = parser.parse_args()
    runner = {
        'run_tf': run_reference_tensorflow,
        'build_trt': build_trt_engine,
        'run_trt': run_trt_engine
    }

    # Run
    runner[args.o]()


